import threading

from cotea.utils import remove_modules_from_imported

# during the imports ansible global objects are created
# that affects to the further work
remove_modules_from_imported(module_name_like="ansible", not_to_delete="ansible_runner")
remove_modules_from_imported(module_name_like="logging")

from ansible.cli import CLI
from ansible.plugins.strategy.linear import StrategyModule
from ansible.plugins.strategy import StrategyBase
from ansible.cli.playbook import PlaybookCLI
from ansible.parsing.yaml.objects import AnsibleUnicode
from ansible.executor.playbook_executor import PlaybookExecutor
from ansible.executor.play_iterator import PlayIterator
from ansible import constants as C
from ansible.playbook.task import Task
from ansible.playbook.helpers import load_list_of_tasks

from cotea.ansible_execution_sync import ans_sync
from cotea.task_result import TaskResult
from cotea.wrappers.pbcli_run_wrapper import pbcli_run_wrapper
from cotea.wrappers.strategy_run_wrapper import strategy_run_wrapper
from cotea.wrappers.get_next_task_wrapper import get_next_task_wrapper
from cotea.wrappers.update_active_conn_wrapper import update_active_conn_wrapper
from cotea.wrappers.play_prereqs_wrapper import play_prereqs_wrapper
from cotea.wrappers.playbook_executor_wrapper import play_executor_wrapper
from cotea.wrappers.iterator_add_task_wrapper import iterator_add_task_wrapper
from cotea.progress_bar import ansible_progress_bar
from cotea.ansible_execution_tree import AnsibleExecTree

import logging



class runner:
    def __init__(self, pb_path, arg_maker, debug_mod=None, show_progress_bar=False):
        logging_lvl = logging.INFO
        if debug_mod:
            logging_lvl= logging.DEBUG
        
        self.show_progress_bar = show_progress_bar
        
        logging.basicConfig(format="%(name)s %(asctime)s %(message)s", \
                    datefmt="%H:%M:%S", level=logging_lvl)

        self.pb_path = pb_path
        self.arg_maker = arg_maker

        self.logger = logging.getLogger("RUNNER")
        
        log_sync = logging.getLogger("SYNC")
        self.sync_obj = ans_sync(log_sync)

        self.breakpoint_labeles = {}
        self.breakpoint_labeles["before_playbook"] = "before_playbook_run"
        self.breakpoint_labeles["after_playbook"] = "after_playbook_run"
        self.breakpoint_labeles["before_play"] = "before_play_run"
        self.breakpoint_labeles["after_play"] = "after_play_run"
        self.breakpoint_labeles["before_task"] = "before_task_run"
        self.breakpoint_labeles["after_task"] = "after_task_run"

        self.progress_bar = ansible_progress_bar()
        self.execution_tree = AnsibleExecTree()

        self._set_wrappers()
        start_ok = self._start_ansible()
        self.logger.debug("Ansible start ok: %s", start_ok)
        

    def _set_wrappers(self):
        wrp_lgr = logging.getLogger("WRPR")

        self.pbcli_run_wrp = pbcli_run_wrapper(PlaybookCLI.run, self.sync_obj, wrp_lgr,
                                               self.breakpoint_labeles["before_playbook"],
                                               self.breakpoint_labeles["after_playbook"])
        PlaybookCLI.run = self.pbcli_run_wrp

        self.play_wrp = strategy_run_wrapper(StrategyModule.run, self.sync_obj, wrp_lgr,
                                             self.breakpoint_labeles["before_play"],
                                             self.breakpoint_labeles["after_play"])
        StrategyModule.run = self.play_wrp

        self.task_wrp = get_next_task_wrapper(StrategyModule._get_next_task_lockstep,
                                              self.sync_obj, wrp_lgr,
                                              self.breakpoint_labeles["before_task"],
                                              self.progress_bar,
                                              self.show_progress_bar)
        StrategyModule._get_next_task_lockstep = self.task_wrp

        self.update_conn_wrapper = update_active_conn_wrapper(StrategyBase.update_active_connections,
                                                              self.sync_obj, wrp_lgr,
                                                              self.breakpoint_labeles["after_task"])
        StrategyBase.update_active_connections = self.update_conn_wrapper

        self.play_prereqs_wrp = play_prereqs_wrapper(CLI._play_prereqs,
                                                      self.sync_obj, wrp_lgr)
        CLI._play_prereqs = self.play_prereqs_wrp

        self.playbook_executor_wrp = play_executor_wrapper(PlaybookExecutor.__init__,
                                                           self.sync_obj, wrp_lgr,
                                                           self.execution_tree,
                                                           self.progress_bar)
        PlaybookExecutor.__init__ = self.playbook_executor_wrp

        self.iterator_add_task_wrp = iterator_add_task_wrapper(PlayIterator.add_tasks,
                                                                  self.sync_obj, wrp_lgr,
                                                                  self.execution_tree,
                                                                  self.progress_bar)
        PlayIterator.add_tasks = self.iterator_add_task_wrp
    

    def _set_wrappers_back(self):
        PlaybookCLI.run = self.pbcli_run_wrp.func
        StrategyModule.run = self.play_wrp.func
        StrategyModule._get_next_task_lockstep = self.task_wrp.func
        StrategyBase.update_active_connections = self.update_conn_wrapper.func
        CLI._play_prereqs = self.play_prereqs_wrp.func
        PlaybookExecutor.__init__ = self.playbook_executor_wrp.func
        PlayIterator.add_tasks = self.iterator_add_task_wrp.func
    

    def _start_ansible(self):
        args = self.arg_maker.args
        args.insert(0, "/usr/local/bin/ansible-playbook")
        args.insert(1, self.pb_path)

        self.pbCLI = PlaybookCLI(args)

        self.ansible_thread = threading.Thread(target=self.pbCLI.run)
        self.ansible_thread.start()
        self.sync_obj.runner_just_wait()

        if self.sync_obj.curr_breakpoint_label == self.breakpoint_labeles["before_playbook"]:
            return True
        
        return False
    

    def has_next_play(self):
        if self.sync_obj.curr_breakpoint_label == self.breakpoint_labeles["after_playbook"]:
            return False
        
        self.sync_obj.continue_ansible_with_stop()
        current_bp_label = self.sync_obj.curr_breakpoint_label
        self.logger.debug("has_next_play: %s", current_bp_label)

        if current_bp_label == self.breakpoint_labeles["before_play"]:
            return True

        return False


    def has_next_task(self):
        if self.sync_obj.curr_breakpoint_label == self.breakpoint_labeles["after_playbook"]:
            return False

        self.sync_obj.continue_ansible_with_stop()
        current_bp_label = self.sync_obj.curr_breakpoint_label
        self.logger.debug("has_next_task: %s", current_bp_label)

        if current_bp_label == self.breakpoint_labeles["before_task"]:
            return True

        return False


    def run_next_task(self):
        res = []

        if self.sync_obj.curr_breakpoint_label == self.breakpoint_labeles["after_playbook"]:
            return res

        self.sync_obj.continue_ansible_with_stop()
        current_bp_label = self.sync_obj.curr_breakpoint_label
        self.logger.debug("run_next_task: %s", current_bp_label)

        if current_bp_label != self.breakpoint_labeles["after_task"]:
            self.logger.debug("run_next_task() has come not in to the 'after_task'")
        
        for task_result_ansible_obj in self.update_conn_wrapper.current_results:
            res.append(TaskResult(task_result_ansible_obj))
        
        return res
    

    def rerun_last_task(self):
        self.task_wrp.rerun_last_task = True
    

    # returns True and empty string if success
    #         False and error msg otherwise
    def add_new_task(self, new_task_str):
        loader = self.playbook_executor_wrp.loader

        ds = None
        try:
            ds = loader.load(new_task_str)
        except Exception as e:
            return False, "Exception during loader.load call\
                (from str to python ds): {}".format(str(e))
        
        if hasattr(ds, "__len__"):
            if len(ds) != 1:
                return False, "You must add 1 new task. Instead you add: {}".format(str(ds))
        else:
            return False, "Python repr of the input string should have\
                        __len__ attr. Maybe something wrong with input: {}\n\
                        Python repr without __len__ attr: {}".format(new_task_str, str(ds))

        current_play = self.play_wrp.iterator._play
        variable_manager = self.playbook_executor_wrp.variable_manager

        try:
            new_ansible_task = load_list_of_tasks(ds=ds, play=current_play, \
                variable_manager=variable_manager, loader=loader)
        except Exception as e:
            return False, "Exception during load_list_of_tasks call\
                (creats Ansible.Task objects): {}".format(str(e))
        
        if hasattr(new_ansible_task, "__len__"):
            new_tasks_count = len(new_ansible_task)
            if new_tasks_count != 1:
                return False, "The input '{}' has been interpreted into {} tasks\
                               instead of 1. Interpretation \
                                result: {}".format(new_task_str, new_tasks_count, str(ds))
        else:
            return False, "Python repr of the input string should have\
                        __len__ attr. Maybe something wrong with input: {}\n\
                        Python repr without __len__ attr: {}".format(new_task_str, str(ds))
        
        # new_task doesn't have parent and parent_type attrs
        # we will take them from previous task
        new_task = new_ansible_task[0]
        ser_new_task = new_task.serialize()
        
        prev_task = self.get_prev_task()
        ser_prev_task = prev_task.serialize()

        if "parent" not in ser_prev_task:
            return False, "Previous task doesn't have 'parent' attr but should.\n\
                Previous task serialize view: {}".format(str(ser_prev_task))

        if hasattr(ser_prev_task["parent"], "copy"):
            ser_new_task["parent"] = ser_prev_task["parent"].copy()
        else:
            return False, "Previous task 'parent' attr doesn't have 'copy' attr but should\
                 because it should be dict. 'parent' attr: {}".format(str(ser_prev_task))

        if "parent_type" not in ser_prev_task:
            return False, "Previous task doesn't have 'parent_type' attr but should.\n\
                Previous task serialize view: {}".format(str(ser_prev_task))
        
        ser_new_task["parent_type"] = ser_prev_task["parent_type"]

        final_new_task = Task()
        final_new_task.deserialize(ser_new_task)

        # needs to be processed properly
        try:
            final_new_task._parent._play = prev_task._parent._play.copy()
        except:
            final_new_task._parent._play = current_play.copy() 
        
        self.task_wrp.new_task_to_add = True
        self.task_wrp.new_task = final_new_task

        return True, ""

    
    def ignore_errors_of_next_task(self):
        self.task_wrp.next_task_ignore_errors = True

    def get_already_ignore_failed(self):
        return self.task_wrp.already_ignore_failed


    def get_already_ignore_unrch(self):
        return self.task_wrp.already_ignore_unrch


    def finish_ansible(self):
        while self.sync_obj.curr_breakpoint_label != self.breakpoint_labeles["after_playbook"]:
            self.sync_obj.continue_ansible_with_stop()
        
        self.sync_obj.continue_ansible()
        self.ansible_thread.join(timeout=5)
        self._set_wrappers_back()
    

    def get_cur_play_name(self):
        return str(self.play_wrp.current_play_name)
    

    def get_next_task(self):
        return self.task_wrp.get_next_task()


    def get_next_task_name(self):
        return str(self.task_wrp.get_next_task_name())
    

    def get_prev_task(self):
        return self.task_wrp.get_prev_task()
    

    def get_prev_task_name(self):
        return str(self.task_wrp.get_prev_task_name())
    

    def get_last_task_result(self):
        res = []

        for task_result_ansible_obj in self.update_conn_wrapper.current_results:
            res.append(TaskResult(task_result_ansible_obj))

        return res
    

    # returns True if there was an non ignored error
    def was_error(self):
        return self.play_wrp.was_error
    

    # returns list with all errors, including the ignored ones
    def get_all_error_msgs(self):
        return self.update_conn_wrapper.error_msgs
    

    # returns last error msg that wasn't ignored 
    def get_error_msg(self):
        res = ""

        # the errors didn't have 'ignore_errors'
        if self.was_error():
            errors_count = len(self.update_conn_wrapper.error_msgs)

            if errors_count > 0:
                res = self.update_conn_wrapper.error_msgs[errors_count - 1]
        
        return res
    

    def get_all_vars(self):
        variable_manager = self.play_wrp.variable_manager
        cur_play = self.play_wrp.iterator._play
        hosts = self.play_wrp.hosts
        hosts_all = self.play_wrp.hosts_all

        res = variable_manager.get_vars(play=cur_play,
                                        _hosts=hosts,
                                        _hosts_all=hosts_all)

        return res


    def get_all_facts(self):
        return self.play_prereqs_wrp.variable_manager._fact_cache.copy()


    def get_variable(self, var_name):
        if var_name == "ansible_facts":
            return self.get_all_facts()

        all_vars = self.get_all_vars()

        if var_name in all_vars:
            return all_vars[var_name]

        # check groups
        if "groups" in all_vars:
            if var_name in all_vars["groups"]:
                return all_vars["groups"][var_name]

        result = {}

        # check hostvars
        if "hostvars" in all_vars:
            for host in all_vars["hostvars"]:
                for key in all_vars["hostvars"][host]:
                    if key == var_name:
                        result[host] = {key: all_vars["hostvars"][host][key]}

        if result:
            return result

        facts = self.get_all_facts()
        for host_key in facts:
            if var_name in facts[host_key]:
                result[host_key] = facts[host_key][var_name]

        if result:
            return result

        self.logger.info("There is no variable with name %s", var_name)

        return None
    

    def add_var_as_extra_var(self, new_var_name, value):
        variable_manager = self.play_wrp.variable_manager

        ansible_way_var = AnsibleUnicode(new_var_name)
        variable_manager._extra_vars[ansible_way_var] = value
   
    
    def _getIP(self):
        var_name = "openstack_servers"
        host_name = "localhost"
        ip1_field_name = "interface_ip"
        ip2_field_name = "private_v4"

        res = ""
        ostack_var = self.get_variable(var_name)

        try:
            if ip1_field_name in ostack_var[host_name][0]:
                res = str(ostack_var[host_name][0][ip1_field_name])
            elif ip2_field_name in ostack_var[host_name][0]:
                res = str(ostack_var[host_name][0][ip2_field_name])
        except Exception as e:
            self.logger.info("During runner._getIP() call error was occured. We skipped it.")
            self.logger.info("Error is:\n%s", e)

        self.logger.debug("get_ip res = %s", res)
        self.logger.debug(type(res))

        return res




from ansible.executor.task_result import TaskResult as TR


class TaskResult:
    def __init__(self, ansible_task_result: TR):
        self.result = ansible_task_result._result.copy()
        self.task_name = str(ansible_task_result.task_name)
        self.task_ansible_object = ansible_task_result._task.copy()
        self.task_fields = ansible_task_result._task_fields.copy()

        self.is_changed = ansible_task_result.is_changed()
        self.is_failed = ansible_task_result.is_failed()
        self.is_skipped = ansible_task_result.is_skipped()
        self.is_unreachable = ansible_task_result.is_unreachable()

        self.stdout = ""
        stdout = self.result.get("stdout")
        if stdout:
            self.stdout = str(stdout)
        
        self.stderr = ""
        stderr = self.result.get("stderr")
        if stderr:
            self.stderr = str(stderr)

        self.msg = ""
        msg = self.result.get("msg")
        if msg:
            self.msg = str(msg)




def remove_modules_from_imported(module_name_like, not_to_delete=None):
    import sys

    for module in list(sys.modules):
        if module_name_like in module:
            if not_to_delete is not None:
                if not_to_delete in module:
                    continue
            
            sys.modules.pop(module)


from ansible.inventory.host import Host

from cotea.wrappers.wrapper_base import wrapper_base
from cotea.wrappers.ansi_breakpoint import ansi_breakpoint
from cotea.progress_bar import ansible_progress_bar

HOST_IND = 0
TASK_IND = 1


# wraps ansible.plugins.strategy.linear.StrategyModule._get_next_task_lockstep()
class get_next_task_wrapper(wrapper_base):
    def __init__(self, func, sync_obj, logger, bp_label, progress_bar: ansible_progress_bar, show_progr_bar=False):
        super().__init__(func, sync_obj, logger)
        self.before_task_bp = ansi_breakpoint(sync_obj, bp_label)
        self.prev_tasks = []
        self.next_tasks = []
        self.show_progress_bar = show_progr_bar

        self.rerun_last_task = False
        self.next_task_ignore_errors = False

        self.new_task_to_add = False
        self.new_task = None

        self.already_ignore_failed = []
        self.already_ignore_unrch = []

        self.progress_bar = progress_bar


    def __call__(self, real_obj, hosts_left, iterator):
        result = None

        new_prev_tasks = self._copy_hosttasks(self.prev_tasks)
        self.prev_tasks = new_prev_tasks

        if self.new_task_to_add:
            next_tasks_copy = self._copy_hosttasks(self.next_tasks)
            next_tasks_with_new_task = self._set_task_in_hosttasks(self.new_task, next_tasks_copy)
            result = next_tasks_with_new_task

        elif self.rerun_last_task:
            result = self.next_tasks

        else:
            new_prev_tasks = self._copy_hosttasks(self.next_tasks)
            self.prev_tasks = new_prev_tasks

            result = self.func(real_obj, hosts_left, iterator)
            self.next_tasks = result

            self.already_ignore_failed = []
            self.already_ignore_unrch = []
                        
        for hosttask in result:
            if hosttask[TASK_IND]:
                # if we rerun the task or add the new one,
                # the list of already ignored tasks should stay the same
                if not self.rerun_last_task and not self.new_task_to_add:
                    if hosttask[TASK_IND].ignore_errors:
                        if hasattr(hosttask[TASK_IND], "get_name"):
                            print(hosttask[TASK_IND].ignore_errors)
                            self.already_ignore_failed.append(str(hosttask[TASK_IND].get_name()))
                    
                    if hosttask[TASK_IND].ignore_unreachable:
                        if hasattr(hosttask[TASK_IND], "get_name"):
                            self.already_ignore_unrch.append(str(hosttask[TASK_IND].get_name()))

                if self.next_task_ignore_errors:
                    hosttask[TASK_IND].ignore_errors = True
                    hosttask[TASK_IND].ignore_unreachable = True

        if self.get_next_task_name() is not None:
            self.progress_bar.update()
            
            if self.show_progress_bar:
                play_name = iterator._play.get_name()
                task_name = None
                if self.new_task_to_add:
                    if hasattr(self.new_task, "get_name"):
                        task_name = self.new_task.get_name()
                    else:
                        task_name = self.get_next_task_name()
                else:
                    task_name = self.get_next_task_name()
                    
                self.progress_bar.print_bar(play_name, task_name)
        
        if self.new_task_to_add:
            self.rerun_last_task = True
        else:
            self.rerun_last_task = False

        self.next_task_ignore_errors = False
        self.new_task_to_add = False
        self.new_task = None

        self.before_task_bp.stop()
        
        '''
        self.logger.info("task run")
        #print(dir(result[0][1]))
        #print(dir(result[0][1]._name))
        if result[0][1] != None:
            if result[0][1].name == "Execute bad command":
                self.zapas_task = result[0][1].copy()
        else:
            if self.zapas_task != None:
                print("aaaaa")
                return [(result[0][0], self.zapas_task)]
        #self.sync_obj.continue_runner_with_stop()

        print("zapas:", type(self.zapas_task))
        '''

        return result

        # все следующие - meta -> прокручиваем их без стопа и без
        #                         возврата управления
        # следующий - норм таск -> возвращаем управление, говоря true
        # следующих не осталось -> возвращаем управление, говоря false


    def _copy_hosttasks(self, hosttasks):
        hosttasks_copy = []

        for hosttask in hosttasks:
            temp_host = None
            temp_task = None

            if hosttask[HOST_IND]:
                temp_host = Host()
                temp_host.deserialize(hosttask[HOST_IND].serialize())
            
            if hosttask[TASK_IND]:
                temp_task = hosttask[TASK_IND].copy()

            if temp_host:
                hosttasks_copy.append( (temp_host, temp_task) )
        
        return hosttasks_copy
    

    def _set_task_in_hosttasks(self, new_task, hosttasks):
        hosttasks_with_new_task = []

        for hosttask in hosttasks:
            temp_host = None

            if hosttask[HOST_IND]:
                temp_host = Host()
                temp_host.deserialize(hosttask[HOST_IND].serialize())

            if temp_host:
                hosttasks_with_new_task.append( (temp_host, new_task) )
        
        return hosttasks_with_new_task


    def get_next_task(self):
        res = None
        
        if len(self.next_tasks) > 0:
            if len(self.next_tasks[0]) >= 2:
                res = self.next_tasks[0][TASK_IND]

        return res
    

    def get_next_task_name(self):
        res = None 
        next_task = self.get_next_task()

        if next_task:
            res = next_task.get_name()
        
        return res
    

    def get_prev_task(self):
        res = None

        if len(self.prev_tasks) > 0:
            if len(self.prev_tasks[0]) >= 2:
                res = self.prev_tasks[0][TASK_IND]

        return res
    
    
    def get_prev_task_name(self):
        res = None
        prev_task = self.get_prev_task()

        if prev_task:
            res = prev_task.get_name()
        
        return res